@@ -34,13 +34,18 @@
 /* for getenv() */
 #include <stdlib.h>
 
-#include "libiptc/libiptc.h"
+#include <string.h>
+
 #include "loader.h"
 
 /* Keep a list of the modules we've loaded */
 static ModuleDef *module_list = NULL;
 
-static void register_module(ModuleDef *, iptc_handle_t *);
+/* Keep an internal reference count; when this is decremented to 0, release
+ * any loaded modules. */
+static int _refcount = 0;
+
+static void register_module(ModuleDef *, HANDLE *, void *);
 
 /* If we have a known built-in target, then we use the 'standard' target
  * module */
@@ -61,16 +66,26 @@ static bool use_std_target(char *targetname) {
 
 /* Load module, using a private call */
 static ModuleDef *find_module_int(char *name, ModuleType type,
-				iptc_handle_t *table, bool dont_load) {
+				HANDLE *table, bool dont_load) {
 	ModuleDef *ptr = NULL;
 	void *libhandle;
 	ModuleDef *(*initf)(void);
+	char *dlname = name;
+
+#ifdef INET6
+	/* What. the. hell. This is an ugly, ugly hack. */
+	if (!strcmp(name, "icmpv6") || !strcmp(name, "ipv6-icmp") ||
+					!strcmp(name, "icmp6")) {
+		dlname = "icmpv6";
+		name = "icmp6";
+	}
+#endif /* INET6 */
 
 	/* If we're looking for a target module, and it's a built-in target or a
 	 * chain, then load the 'standard' target module instead... */
 	if((type == MODULE_TARGET) && (use_std_target(name) ||
-							(table && iptc_is_chain(name, *table))))
-		name = STD_TARGET;
+							(table && IS_CHAIN(name, *table))))
+		dlname = name = STD_TARGET;
 
 	/* If the module we need has already been loaded, then just kick back
 	 * a pointer to its data structure */
@@ -84,29 +99,61 @@ static ModuleDef *find_module_int(char *name, ModuleType type,
 	 * itself */
 	if(!dont_load) {
 		char *path, *basepath = NULL;
-		basepath = getenv("IPT_IPV4_MODPATH");
+		basepath = getenv("IPT_MODPATH");
 		if (!basepath || !strcmp(basepath, ""))
 			basepath = MODULE_PATH;
-		asprintf(&path, "%s/ipt_pl_%s.so", basepath, name);
+#ifdef INET6
+		asprintf(&path, "%s/ip6t_pl_%s.so", basepath, dlname);
+#else /* !INET6 */
+		asprintf(&path, "%s/ipt_pl_%s.so", basepath, dlname);
+#endif /* INET6 */
 		if((libhandle = dlopen(path, RTLD_NOW))) {
 			initf = dlsym(libhandle, "init");
-			register_module(initf(), table);
+			register_module(initf(), table, libhandle);
 			if(!(ptr = find_module_int(name, type, table, TRUE)))
-				fprintf(stderr, "Couldn't load module '%s'\n", name);
-		}
+				SET_ERRSTR("Couldn't lookup module %s after registration", name);
+		} else
+			SET_ERRSTR("dlopen() failed: %s", dlerror());
 		free(path);
 	}
 	return(ptr);
 }
 
 /* Look up, or load, a module for me */
-ModuleDef *ipt_find_module(char *name, ModuleType type, iptc_handle_t *table) {
+ModuleDef *ipt_find_module(char *name, ModuleType type, HANDLE *table) {
 	return find_module_int(name, type, table, FALSE);
 }
 
+/* Keep an internal count of how many times we've been instantiated, so that
+ * we can unload the modules we've loaded */
+void ipt_loader_setup(void) {
+	_refcount++;
+}
+
+/* Release all loaded modules when the refcount is 0 */
+void ipt_release_modules(void) {
+	ModuleDef *next;
+	--_refcount;
+	if (_refcount < 0)
+		printf("refcount less than 0, wtf?\n");
+	if (_refcount)
+		return;
+	/* This way, if someone really doesn't want to unload the modules, they
+	 * can say so. This is useful with tools like valgrind and gdb. */
+	if (!getenv("IPT_DONT_UNLOAD")) {
+		while (module_list) {
+			next = module_list->next;
+			dlclose(module_list->libh);
+			module_list = next;
+		}
+	}
+}
+
 /* Indoctrinate the newly-loaded module into the ways of our lands... */
-static void register_module(ModuleDef *def, iptc_handle_t *table) {
+static void register_module(ModuleDef *def, HANDLE *table,
+				void *libhandle) {
 	ModuleDef *i;
+	def->libh = libhandle;
 	/* Make sure this module isn't already loaded - if it is, something is
 	 * broken. This should never happen, but if it does, then we probably 
 	 * have a faulty module. */
@@ -117,7 +164,7 @@ static void register_module(ModuleDef *def, iptc_handle_t *table) {
 	}
 
 	/* Check the size alignment - if alignment is wrong, we've got problems */
-	if(def->size != IPT_ALIGN(def->size)) {
+	if(def->size != ALIGN(def->size)) {
 		fprintf(stderr, "Size is not properly aligned for this "
 						"architecture!\n");
 		exit(1);
